"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("./util");
const entities_1 = require("./entities");
const node_fetch_1 = __importDefault(require("./node-fetch"));
/** @internal */
const DEFAULT_CONFIG = {
    logLevel: 1,
    dbURL: 'http://boomlings.com/database',
    corsURL: 'https://cors-anywhere.herokuapp.com/',
    fetch: node_fetch_1.default
};
/**
 * Client for Geometry Dash requests.
 */
class Client {
    /**
     * Creates a client for Geometry Dash requests.
     * @param config The configuration for the client.
     */
    constructor(config) {
        this.config = {
            ...DEFAULT_CONFIG,
            ...config
        };
        this.users = new entities_1.UserCreator(this);
        this.levels = new entities_1.LevelCreator(this);
    }
    /**
     * Make a request to a Geometry Dash server.
     *
     * @param url The path to request to (based at the {@link Config.dbURL})
     * @param conf The request configuration
     * @param returnRaw Whether to parse the response into a string or return it raw
     * @return The Response or string containing the Geometry Dash server's response
     * @internal
     */
    async req(url, { method = 'GET', body = null } = {}, returnRaw = false) {
        let sentBody = null;
        if (body) {
            sentBody = body.resolve();
        }
        const resp = await node_fetch_1.default((util_1.isNode ? '' : this.config.corsURL) + (url.startsWith('http') ? '' : this.config.dbURL) + url, {
            method,
            referrerPolicy: 'no-referrer',
            body: sentBody
        });
        if (returnRaw)
            return resp;
        const data = await resp.text();
        this.verbose(`Made a ${method} request to ${url}, response: ${data}`);
        return data;
    }
    /** @internal */
    warn(str) {
        if (this.config.logLevel > 0)
            console.warn(str);
    }
    /** @internal */
    log(str) {
        if (this.config.logLevel > 0)
            console.log(str);
    }
    /** @internal */
    verbose(str) {
        if (this.config.logLevel > 1)
            console.log(str);
    }
}
exports.default = Client;
